@@ -29,10 +29,11 @@ import java.util.concurrent.Future;
 import org.apache.commons.lang.NotImplementedException;
 import org.apache.commons.logging.Log;
 import org.apache.commons.logging.LogFactory;
-import org.apache.sysds.hops.OptimizerUtils;
 import org.apache.sysds.runtime.DMLRuntimeException;
 import org.apache.sysds.runtime.compress.CompressedMatrixBlock;
+import org.apache.sysds.runtime.compress.CompressedMatrixBlockFactory;
 import org.apache.sysds.runtime.compress.CompressionSettings;
+import org.apache.sysds.runtime.compress.DMLCompressionException;
 import org.apache.sysds.runtime.compress.colgroup.AColGroup;
 import org.apache.sysds.runtime.compress.colgroup.ColGroupConst;
 import org.apache.sysds.runtime.compress.colgroup.dictionary.ADictionary;
@@ -40,6 +41,7 @@ import org.apache.sysds.runtime.compress.colgroup.dictionary.MatrixBlockDictiona
 import org.apache.sysds.runtime.data.SparseBlock;
 import org.apache.sysds.runtime.functionobjects.Divide;
 import org.apache.sysds.runtime.functionobjects.Minus;
+import org.apache.sysds.runtime.functionobjects.Minus1Multiply;
 import org.apache.sysds.runtime.functionobjects.MinusMultiply;
 import org.apache.sysds.runtime.functionobjects.Multiply;
 import org.apache.sysds.runtime.functionobjects.Plus;
@@ -59,42 +61,101 @@ public class CLALibBinaryCellOp {
 
 	private static final Log LOG = LogFactory.getLog(CLALibBinaryCellOp.class.getName());
 
-	public static MatrixBlock binaryOperations(BinaryOperator op, CompressedMatrixBlock m1, MatrixValue thatValue,
-		MatrixValue result) {
-		MatrixBlock that = CompressedMatrixBlock.getUncompressed(thatValue);
-		LibMatrixBincell.isValidDimensionsBinary(m1, that);
-		BinaryAccessType atype = LibMatrixBincell.getBinaryAccessType(m1, that);
-		return selectProcessingBasedOnAccessType(op, m1, that, thatValue, result, atype, false);
+	public static MatrixBlock binaryOperations(BinaryOperator op, CompressedMatrixBlock m1, MatrixBlock that,
+		MatrixBlock result) {
+		try {
+			if(that.getNumRows() == 1 && that.getNumColumns() == 1) {
+				ScalarOperator sop = new RightScalarOperator(op.fn, that.getValue(0, 0), op.getNumThreads());
+				return CLALibScalar.scalarOperations(sop, m1, result);
+			}
+			if(that.isEmpty())
+				return binaryOperationsEmpty(op, m1, that, result);
+			that = CompressedMatrixBlock.getUncompressed(that, "Decompressing right side in BinaryOps");
+			LibMatrixBincell.isValidDimensionsBinary(m1, that);
+			BinaryAccessType atype = LibMatrixBincell.getBinaryAccessType(m1, that);
+			return selectProcessingBasedOnAccessType(op, m1, that, result, atype, false);
+		}
+		catch(Exception e) {
+			throw new DMLCompressionException("Failed to perform compressed binary operation: " + op, e);
+		}
 	}
 
-	public static MatrixBlock binaryOperationsLeft(BinaryOperator op, CompressedMatrixBlock m1, MatrixValue thatValue,
-		MatrixValue result) {
-		MatrixBlock that = CompressedMatrixBlock.getUncompressed(thatValue);
+	public static MatrixBlock binaryOperationsLeft(BinaryOperator op, CompressedMatrixBlock m1, MatrixBlock that,
+		MatrixBlock result) {
+		if(that.getNumRows() == 1 && that.getNumColumns() == 1) {
+			ScalarOperator sop = new LeftScalarOperator(op.fn, that.getValue(0, 0), op.getNumThreads());
+			return CLALibScalar.scalarOperations(sop, m1, result);
+		}
+		if(that.isEmpty())
+			throw new NotImplementedException("Not handling left empty yet");
+
+		that = CompressedMatrixBlock.getUncompressed(that, "Decompressing left side in BinaryOps");
 		LibMatrixBincell.isValidDimensionsBinary(that, m1);
 		BinaryAccessType atype = LibMatrixBincell.getBinaryAccessType(that, m1);
-		return selectProcessingBasedOnAccessType(op, m1, that, thatValue, result, atype, true);
+		return selectProcessingBasedOnAccessType(op, m1, that, result, atype, true);
+	}
+
+	private static MatrixBlock binaryOperationsEmpty(BinaryOperator op, CompressedMatrixBlock m1, MatrixBlock that,
+		MatrixBlock result) {
+		final ValueFunction fn = op.fn;
+		if((m1.getNumRows() == that.getNumRows() && m1.getNumColumns() == that.getNumColumns()) ||
+			m1.getNumColumns() == that.getNumColumns()) {
+
+			if(fn instanceof Multiply)
+				result = CompressedMatrixBlockFactory.createConstant(m1.getNumRows(), m1.getNumColumns(), 0);
+			else if(fn instanceof Minus1Multiply)
+				result = CompressedMatrixBlockFactory.createConstant(m1.getNumRows(), m1.getNumColumns(), 1);
+			else if(fn instanceof Minus || fn instanceof Plus || fn instanceof MinusMultiply ||
+				fn instanceof PlusMultiply) {
+				CompressedMatrixBlock ret = new CompressedMatrixBlock();
+				ret.copy(m1);
+				return ret;
+			}
+			else
+				throw new NotImplementedException("Function Type: " + fn);
+
+			return result;
+		}
+		else {
+			final long lr = m1.getNumRows();
+			final long rr = that.getNumRows();
+			final long lc = m1.getNumColumns();
+			final long rc = that.getNumColumns();
+			throw new NotImplementedException(
+				"Not Implemented sizes: left(" + lr + ", " + lc + ") right(" + rr + ", " + rc + ")");
+		}
 	}
 
 	private static MatrixBlock selectProcessingBasedOnAccessType(BinaryOperator op, CompressedMatrixBlock m1,
-		MatrixBlock that, MatrixValue thatValue, MatrixValue result, BinaryAccessType atype, boolean left) {
-		if(atype == BinaryAccessType.MATRIX_COL_VECTOR)
-			return binaryMVCol(m1, that, op, left);
-		else if(atype == BinaryAccessType.MATRIX_MATRIX) {
-			if(that.isEmpty()) {
-				ScalarOperator sop = left ? new LeftScalarOperator(op.fn, 0, -1) : new RightScalarOperator(op.fn, 0,
-					-1);
-				return CLALibScalar.scalarOperations(sop, m1, result);
+		MatrixBlock that, MatrixBlock result, BinaryAccessType atype, boolean left) {
+
+		if(atype == BinaryAccessType.MATRIX_COL_VECTOR) {
+
+			MatrixBlock d_compressed = m1.getCachedDecompressed();
+
+			if(d_compressed != null){
+				if(left)
+					return that.binaryOperations(op, d_compressed);
+				else
+					return d_compressed.binaryOperations(op, that);
 			}
-			else if(that.isInSparseFormat())
-				return binaryMMSparse(m1, that, op, left);
+
+			return binaryMVCol(m1, that, op, left);
+		}
+		else if( atype == BinaryAccessType.MATRIX_MATRIX) {
+
+			MatrixBlock d_compressed = m1.getUncompressed("MatrixMatrix " + op);
+
+			if(left)
+				return that.binaryOperations(op, d_compressed);
 			else
-				return binaryMMDense(m1, that, op, left);
+				return d_compressed.binaryOperations(op, that);
+
 		}
 		else if(isSupportedBinaryCellOp(op.fn))
 			return bincellOp(m1, that, setupCompressedReturnMatrixBlock(m1, result), op, left);
 		else {
-			LOG.warn("Decompressing since Binary Ops" + op.fn + " is not supported compressed");
-			return CompressedMatrixBlock.getUncompressed(m1).binaryOperations(op, thatValue, result);
+			return CompressedMatrixBlock.getUncompressed(m1, "BinaryOp: " + op.fn).binaryOperations(op, that, result);
 		}
 	}
 
@@ -252,7 +313,7 @@ public class CLALibBinaryCellOp {
 			for(int i = 0; i < nCol; i++)
 				colIndexes[i] = i;
 			ADictionary newDict = new MatrixBlockDictionary(m2);
-			newColGroups.add(new ColGroupConst(colIndexes, m1.getNumRows(), newDict));
+			newColGroups.add(new ColGroupConst(colIndexes, newDict));
 		}
 		else {
 			AColGroup g = newColGroups.get(smallestIndex).binaryRowOp(op, m2.getDenseBlockValues(), false, left);
@@ -269,72 +330,46 @@ public class CLALibBinaryCellOp {
 
 		MatrixBlock ret = new MatrixBlock(m1.getNumRows(), m1.getNumColumns(), false, -1).allocateBlock();
 
-		final int blkz = CompressionSettings.BITMAP_BLOCK_SZ;
-		int k = OptimizerUtils.getConstrainedNumThreads(-1);
-		ExecutorService pool = CommonThreadPool.get(k);
-		ArrayList<Callable<Integer>> tasks = new ArrayList<>();
+		final int blkz = CompressionSettings.BITMAP_BLOCK_SZ / m1.getNumColumns() * 5;
+		final int k = op.getNumThreads();
+		long nnz = 0;
 
-		try {
+		if(k <= 1) {
 			for(int i = 0; i * blkz < m1.getNumRows(); i++) {
 				if(left)
-					tasks.add(
-						new BinaryMVColLeftTask(m1, m2, ret, i * blkz, Math.min(m1.getNumRows(), (i + 1) * blkz), op));
+					nnz += new BinaryMVColLeftTask(m1, m2, ret, i * blkz, Math.min(m1.getNumRows(), (i + 1) * blkz), op)
+						.call();
 				else
-					tasks
-						.add(new BinaryMVColTask(m1, m2, ret, i * blkz, Math.min(m1.getNumRows(), (i + 1) * blkz), op));
+					nnz += new BinaryMVColTask(m1, m2, ret, i * blkz, Math.min(m1.getNumRows(), (i + 1) * blkz), op)
+						.call();
 
 			}
-			long nnz = 0;
-			for(Future<Integer> f : pool.invokeAll(tasks))
-				nnz += f.get();
-			ret.setNonZeros(nnz);
-			pool.shutdown();
-		}
-		catch(InterruptedException | ExecutionException e) {
-			e.printStackTrace();
-			throw new DMLRuntimeException(e);
 		}
+		else {
+			ExecutorService pool = CommonThreadPool.get(op.getNumThreads());
+			ArrayList<Callable<Integer>> tasks = new ArrayList<>();
+			try {
+				for(int i = 0; i * blkz < m1.getNumRows(); i++) {
+					if(left)
+						tasks.add(new BinaryMVColLeftTask(m1, m2, ret, i * blkz,
+							Math.min(m1.getNumRows(), (i + 1) * blkz), op));
+					else
+						tasks.add(
+							new BinaryMVColTask(m1, m2, ret, i * blkz, Math.min(m1.getNumRows(), (i + 1) * blkz), op));
 
-		return ret;
-	}
-
-	private static MatrixBlock binaryMMDense(CompressedMatrixBlock m1, MatrixBlock m2, BinaryOperator op,
-		boolean left) {
-
-		MatrixBlock ret = new MatrixBlock(m1.getNumRows(), m1.getNumColumns(), false, -1).allocateBlock();
-
-		// final int blkz = CompressionSettings.BITMAP_BLOCK_SZ * 2 / m1.getNumColumns();
-
-		final int blkz = CompressionSettings.BITMAP_BLOCK_SZ * 6 / m1.getNumColumns();
-		int k = OptimizerUtils.getConstrainedNumThreads(-1);
-		ExecutorService pool = CommonThreadPool.get(k);
-		ArrayList<Callable<Integer>> tasks = new ArrayList<>();
-
-		try {
-			for(int i = 0; i * blkz < m1.getNumRows(); i++) {
-				int start = i * blkz;
-				int end = Math.min(m1.getNumRows(), (i + 1) * blkz);
-				if(left)
-					tasks.add(new BinaryMMLeftTask(m1, m2, ret, start, end, op));
-				else
-					tasks.add(new BinaryMMTask(m1, m2, ret, start, end, op));
+				}
+				for(Future<Integer> f : pool.invokeAll(tasks))
+					nnz += f.get();
+				pool.shutdown();
+			}
+			catch(InterruptedException | ExecutionException e) {
+				e.printStackTrace();
+				throw new DMLRuntimeException(e);
 			}
-			long nnz = 0;
-			for(Future<Integer> f : pool.invokeAll(tasks))
-				nnz += f.get();
-			ret.setNonZeros(nnz);
-			pool.shutdown();
 		}
-		catch(InterruptedException | ExecutionException e) {
-			e.printStackTrace();
-			throw new DMLRuntimeException(e);
-		}
-		return ret;
-	}
+		ret.setNonZeros(nnz);
 
-	private static MatrixBlock binaryMMSparse(CompressedMatrixBlock m1, MatrixBlock m2, BinaryOperator op,
-		boolean left) {
-		throw new NotImplementedException("not implemented sparse Binary MM " + op.fn);
+		return ret;
 	}
 
 	private static class BinaryMVColTask implements Callable<Integer> {
@@ -359,7 +394,7 @@ public class CLALibBinaryCellOp {
 		public Integer call() {
 			// unsafe decompress, since we count nonzeros afterwards.
 			for(AColGroup g : _m1.getColGroups())
-				g.decompressToBlockUnSafe(_ret, _rl, _ru);
+				g.decompressToBlock(_ret, _rl, _ru);
 
 			if(_m2.isInSparseFormat())
 				throw new NotImplementedException("Not Implemented sparse Format execution for MM.");
@@ -403,7 +438,7 @@ public class CLALibBinaryCellOp {
 		public Integer call() {
 			// unsafe decompress, since we count nonzeros afterwards.
 			for(AColGroup g : _m1.getColGroups())
-				g.decompressToBlockUnSafe(_ret, _rl, _ru);
+				g.decompressToBlock(_ret, _rl, _ru);
 
 			if(_m2.isInSparseFormat())
 				throw new NotImplementedException("Not Implemented sparse Format execution for MM.");
@@ -425,96 +460,6 @@ public class CLALibBinaryCellOp {
 		}
 	}
 
-	private static class BinaryMMTask implements Callable<Integer> {
-		private final int _rl;
-		private final int _ru;
-		private final CompressedMatrixBlock _m1;
-		private final MatrixBlock _m2;
-		private final MatrixBlock _ret;
-		private final BinaryOperator _op;
-
-		protected BinaryMMTask(CompressedMatrixBlock m1, MatrixBlock m2, MatrixBlock ret, int rl, int ru,
-			BinaryOperator op) {
-			_m1 = m1;
-			_m2 = m2;
-			_ret = ret;
-			_op = op;
-			_rl = rl;
-			_ru = ru;
-		}
-
-		@Override
-		public Integer call() {
-			AColGroup.decompressColumnToBlockUnSafe(_ret, _rl, _ru, _m1.getColGroups());
-
-			if(_m2.isInSparseFormat()) {
-				throw new NotImplementedException("Not Implemented sparse Format execution for MM.");
-			}
-			else {
-
-				double[] _retDense = _ret.getDenseBlockValues();
-				double[] _m2Dense = _m2.getDenseBlockValues();
-				int nnz = 0;
-				int offset = _rl * _m1.getNumColumns();
-				for(int row = _rl; row < _ru; row++) {
-					for(int col = 0; col < _m1.getNumColumns(); col++) {
-						double v = _op.fn.execute(_retDense[offset], _m2Dense[offset]);
-						nnz += (v != 0) ? 1 : 0;
-						_retDense[offset] = v;
-						offset++;
-					}
-				}
-				return nnz;
-			}
-		}
-	}
-
-	private static class BinaryMMLeftTask implements Callable<Integer> {
-		private final int _rl;
-		private final int _ru;
-		private final CompressedMatrixBlock _m1;
-		private final MatrixBlock _m2;
-		private final MatrixBlock _ret;
-		private final BinaryOperator _op;
-
-		protected BinaryMMLeftTask(CompressedMatrixBlock m1, MatrixBlock m2, MatrixBlock ret, int rl, int ru,
-			BinaryOperator op) {
-			_m1 = m1;
-			_m2 = m2;
-			_ret = ret;
-			_op = op;
-			_rl = rl;
-			_ru = ru;
-		}
-
-		@Override
-		public Integer call() {
-			AColGroup.decompressColumnToBlockUnSafe(_ret, _rl, _ru, _m1.getColGroups());
-
-			if(_m2.isInSparseFormat()) {
-				throw new NotImplementedException("Not Implemented sparse Format execution for MM.");
-			}
-			else {
-
-				double[] _retDense = _ret.getDenseBlockValues();
-				double[] _m2Dense = _m2.getDenseBlockValues();
-				int nnz = 0;
-				int numCols = _m1.getNumColumns();
-				int offset = _rl * numCols;
-				for(int row = _rl; row < _ru; row++) {
-					for(int col = 0; col < numCols; col++) {
-						double v = _op.fn.execute(_m2Dense[offset], _retDense[offset]);
-						nnz += (v != 0) ? 1 : 0;
-						_retDense[offset] = v;
-						offset++;
-					}
-				}
-				return nnz;
-			}
-
-		}
-	}
-
 	private static class BinaryMVRowTask implements Callable<AColGroup> {
 		private final AColGroup _group;
 		private final double[] _v;
