@@ -38,6 +38,7 @@ import org.apache.sysds.common.Types.ValueType;
 import org.apache.sysds.conf.ConfigurationManager;
 import org.apache.sysds.hops.OptimizerUtils;
 import org.apache.sysds.runtime.DMLRuntimeException;
+import org.apache.sysds.runtime.compress.CompressedMatrixBlock;
 import org.apache.sysds.runtime.controlprogram.caching.LazyWriteBuffer.RPolicy;
 import org.apache.sysds.runtime.controlprogram.federated.FederationMap;
 import org.apache.sysds.runtime.controlprogram.federated.FederationMap.FType;
@@ -178,6 +179,10 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 	protected MetaData _metaData = null;
 	
 	protected FederationMap _fedMapping = null;
+
+	protected boolean _compressed = false;
+
+	protected long _compressedSize = -1;
 	
 	/** The name of HDFS file in which the data is backed up. */
 	protected String _hdfsFileName = null; // file name and path
@@ -243,8 +248,10 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		_gpuObjects = that._gpuObjects;
 		_privacyConstraint = that._privacyConstraint;
 		_dirtyFlag = that._dirtyFlag;
+		_compressed = that._compressed;
+		_compressedSize = that._compressedSize;
+		_fedMapping = that._fedMapping;
 	}
-
 	
 	/**
 	 * Enables or disables the cleanup of the associated 
@@ -320,7 +327,20 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 	public void setMetaData(MetaData md) {
 		_metaData = md;
 	}
+
+	public void setCompressedSize(long size){
+		_compressed = true;
+		_compressedSize = size;
+	}
+
+	public boolean isCompressed(){
+		return _compressed;
+	}
 	
+	public long getCompressedSize(){
+		return _compressedSize;
+	}
+
 	@Override
 	public MetaData getMetaData() {
 		return _metaData;
@@ -381,6 +401,10 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		return isFederated() && (type == null || _fedMapping.getType().isType(type));
 	}
 	
+	public boolean isFederatedExcept(FType type) {
+		return isFederated() && !isFederated(type);
+	}
+	
 	/**
 	 * Gets the mapping of indices ranges to federated objects.
 	 * @return fedMapping mapping
@@ -523,7 +547,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 					if( DMLScript.STATISTICS )
 						CacheStatistics.incrementLinHits();
 				}
-				else if( isFederated() ) {
+				else if( isFederatedExcept(FType.BROADCAST) ) {
 					_data = readBlobFromFederated(_fedMapping);
 
 					//mark for initial local write despite read operation
@@ -598,6 +622,10 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 				Statistics.addCPMemObject(System.identityHashCode(this), getDataSize());
 		}
 		
+		if(newData instanceof CompressedMatrixBlock) {
+			setCompressedSize(newData.getInMemorySize());
+		}
+
 		return ret;
 	}
 	
@@ -761,7 +789,6 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 	 */
 	public synchronized void exportData( int replication ) {
 		exportData(_hdfsFileName, null, replication, null);
-		_hdfsFileExists = true;
 	}
 
 	public synchronized void exportData(String fName, String outputFormat) {
@@ -772,10 +799,6 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		exportData(fName, outputFormat, -1, formatProperties);
 	}
 	
-	public synchronized void exportData (String fName, String outputFormat, int replication, FileFormatProperties formatProperties) {
-		exportData(fName, outputFormat, replication, formatProperties, null);
-	}
-	
 	/**
 	 * Synchronized because there might be parallel threads (parfor local) that
 	 * access the same object (in case it was created before the loop).
@@ -791,9 +814,8 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 	 * @param outputFormat format
 	 * @param replication ?
 	 * @param formatProperties file format properties
-	 * @param opcode instruction opcode if available
 	 */
-	public synchronized void exportData (String fName, String outputFormat, int replication, FileFormatProperties formatProperties, String opcode) {
+	public synchronized void exportData (String fName, String outputFormat, int replication, FileFormatProperties formatProperties) {
 		if( LOG.isTraceEnabled() )
 			LOG.trace("Export data "+hashCode()+" "+fName);
 		long t0 = DMLScript.STATISTICS ? System.nanoTime() : 0;
@@ -823,11 +845,13 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 			setHDFSFileExists(true);
 		
 		//check for common file scheme (otherwise no copy/rename)
+		int blen = (formatProperties == null) ?
+			ConfigurationManager.getBlocksize() : formatProperties.getBlocksize();
 		boolean eqScheme = IOUtilFunctions.isSameFileScheme(
 			new Path(_hdfsFileName), new Path(fName));
 		boolean eqFormat = isEqualOutputFormat(outputFormat);
-		boolean eqBlksize = (outputFormat == null || outputFormat.equals("binary"))
-			&& ConfigurationManager.getBlocksize() != getBlocksize();
+		boolean eqBlksize = (getBlocksize() != blen)
+			&& (outputFormat == null || outputFormat.equals("binary"));
 		
 		//actual export (note: no direct transfer of local copy in order to ensure blocking (and hence, parallelism))
 		if( isDirty() || !eqScheme || isFederated() ||
@@ -917,7 +941,8 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 			//CASE 4: data already in hdfs (do nothing, no need for export)
 			LOG.trace(this.getDebugName() + ": Skip export to hdfs since data already exists.");
 		}
-		  
+		
+		_hdfsFileExists = true;
 		if( DMLScript.STATISTICS ){
 			long t1 = System.nanoTime();
 			CacheStatistics.incrementExportTime(t1-t0);
@@ -1066,6 +1091,8 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		if ( fmt != FileFormat.MM ) {
 			// Get the dimension information from the metadata stored within MatrixObject
 			DataCharacteristics dc = iimd.getDataCharacteristics();
+			if( formatProperties != null && formatProperties.knownBlocksize() )
+				dc.setBlocksize(formatProperties.getBlocksize());
 			
 			// when outputFormat is binaryblock, make sure that matrixCharacteristics has correct blocking dimensions
 			// note: this is only required if singlenode (due to binarycell default) 
@@ -1214,6 +1241,10 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		return (_cacheStatus == CacheStatus.MODIFY);
 	}
 	
+	public boolean isPendingRDDOps() {
+		return isEmpty(true) && _data == null && (_rddHandle != null && _rddHandle.hasBackReference());
+	}
+	
 	protected void setEmpty() {
 		_cacheStatus = CacheStatus.EMPTY;
 	}
@@ -1452,11 +1483,14 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 				str.append("null, null");
 			}
 		} catch (Exception ex) {
-			LOG.error(ex);
+			throw new DMLRuntimeException(ex);
 		}
 		str.append(", ");
 		str.append(isDirty() ? "dirty" : "not-dirty");
 
+		if(isCompressed())
+			str.append( ", Compressed " + _compressedSize );
+
 		return str.toString();
 	}
 }
