@@ -19,23 +19,28 @@
 
 package org.apache.sysds.test;
 
-import static java.lang.Thread.sleep;
-import static org.junit.Assert.assertEquals;
-import static org.junit.Assert.fail;
-
 import java.io.ByteArrayOutputStream;
 import java.io.File;
 import java.io.IOException;
 import java.io.PrintStream;
+import java.net.InetSocketAddress;
 import java.net.ServerSocket;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.HashMap;
+import java.util.List;
 import java.util.Map;
 import java.util.Properties;
+import java.util.Set;
+import java.util.stream.Collectors;
 
+import static java.lang.Math.ceil;
+import static java.lang.Thread.sleep;
+import static org.junit.Assert.assertEquals;
+import static org.junit.Assert.fail;
 import org.apache.commons.io.FileUtils;
 import org.apache.commons.io.IOUtils;
+import org.apache.commons.lang3.tuple.Pair;
 import org.apache.commons.logging.Log;
 import org.apache.commons.logging.LogFactory;
 import org.apache.hadoop.conf.Configuration;
@@ -43,36 +48,48 @@ import org.apache.hadoop.util.GenericOptionsParser;
 import org.apache.spark.sql.SparkSession;
 import org.apache.spark.sql.SparkSession.Builder;
 import org.apache.sysds.api.DMLScript;
+import org.apache.sysds.common.Types;
 import org.apache.sysds.common.Types.DataType;
 import org.apache.sysds.common.Types.ExecMode;
+import org.apache.sysds.common.Types.ExecType;
 import org.apache.sysds.common.Types.FileFormat;
 import org.apache.sysds.common.Types.ValueType;
 import org.apache.sysds.conf.ConfigurationManager;
 import org.apache.sysds.conf.DMLConfig;
 import org.apache.sysds.hops.OptimizerUtils;
 import org.apache.sysds.lops.Lop;
-import org.apache.sysds.lops.LopProperties.ExecType;
+import org.apache.sysds.lops.compile.Dag;
 import org.apache.sysds.parser.DataExpression;
 import org.apache.sysds.parser.ParseException;
 import org.apache.sysds.runtime.DMLRuntimeException;
 import org.apache.sysds.runtime.DMLScriptException;
+import org.apache.sysds.runtime.controlprogram.caching.MatrixObject;
 import org.apache.sysds.runtime.controlprogram.context.SparkExecutionContext;
+import org.apache.sysds.runtime.controlprogram.federated.FederatedData;
+import org.apache.sysds.runtime.controlprogram.federated.FederatedRange;
+import org.apache.sysds.runtime.controlprogram.federated.FederationMap;
+import org.apache.sysds.runtime.controlprogram.federated.FederationUtils;
+import org.apache.sysds.runtime.io.FileFormatProperties;
 import org.apache.sysds.runtime.io.FileFormatPropertiesCSV;
 import org.apache.sysds.runtime.io.FrameReader;
 import org.apache.sysds.runtime.io.FrameReaderFactory;
+import org.apache.sysds.runtime.io.ReaderWriterFederated;
 import org.apache.sysds.runtime.matrix.data.FrameBlock;
 import org.apache.sysds.runtime.matrix.data.MatrixBlock;
+import org.apache.sysds.runtime.matrix.data.MatrixValue;
 import org.apache.sysds.runtime.matrix.data.MatrixValue.CellIndex;
 import org.apache.sysds.runtime.meta.MatrixCharacteristics;
+import org.apache.sysds.runtime.meta.MetaDataAll;
+import org.apache.sysds.runtime.meta.MetaDataFormat;
 import org.apache.sysds.runtime.privacy.CheckedConstraintsLog;
 import org.apache.sysds.runtime.privacy.PrivacyConstraint;
 import org.apache.sysds.runtime.privacy.PrivacyConstraint.PrivacyLevel;
+import org.apache.sysds.runtime.privacy.PrivacyUtils;
 import org.apache.sysds.runtime.util.DataConverter;
 import org.apache.sysds.runtime.util.HDFSTool;
 import org.apache.sysds.utils.ParameterBuilder;
 import org.apache.sysds.utils.Statistics;
 import org.apache.wink.json4j.JSONException;
-import org.apache.wink.json4j.JSONObject;
 import org.junit.After;
 import org.junit.Assert;
 import org.junit.Before;
@@ -90,7 +107,6 @@ import org.junit.Before;
  * </ul>
  *
  */
-@SuppressWarnings("deprecation")
 public abstract class AutomatedTestBase {
 
 	private static final Log LOG = LogFactory.getLog(AutomatedTestBase.class.getName());
@@ -103,6 +119,8 @@ public abstract class AutomatedTestBase {
 	public static final double GPU_TOLERANCE = 1e-9;
 
 	public static final int FED_WORKER_WAIT = 1000; // in ms
+	public static final int FED_WORKER_WAIT_S = 50; // in ms
+	
 
 	// With OpenJDK 8u242 on Windows, the new changes in JDK are not allowing
 	// to set the native library paths internally thus breaking the code.
@@ -113,6 +131,7 @@ public abstract class AutomatedTestBase {
 	 * Script source directory for .dml and .r files only (TEST_DATA_DIR for generated test data artifacts).
 	 */
 	protected static final String SCRIPT_DIR = "./src/test/scripts/";
+	protected static final String DATASET_DIR = "./src/test/resources/datasets/";
 	protected static final String INPUT_DIR = "in/";
 	protected static final String OUTPUT_DIR = "out/";
 	protected static final String EXPECTED_DIR = "expected/";
@@ -127,6 +146,7 @@ public abstract class AutomatedTestBase {
 	 * Location of the SystemDS config file that we use as a template when generating the configs for each test case.
 	 */
 	private static final File CONFIG_TEMPLATE_FILE = new File(CONFIG_DIR, "SystemDS-config.xml");
+	protected boolean disableConfigFile = false;
 
 	protected enum CodegenTestType {
 		DEFAULT, FUSE_ALL, FUSE_NO_REDUNDANCY;
@@ -200,19 +220,36 @@ public abstract class AutomatedTestBase {
 	private boolean isOutAndExpectedDeletionDisabled = false;
 
 	private static boolean outputBuffering = false;
-	
+
 	static {
-		java.io.InputStream inputStream = Thread.currentThread().getContextClassLoader().getResourceAsStream("my.properties");
+		// Load configuration from setting file build by maven.
+		// If maven is not used as test setup, (as default in intellij for instance) default values are used.
+		// If one wants to use custom configurations, setup the IDE to build using maven, and set execution flags
+		// accordingly.
+		// Settings available can be found in the properties inside pom.xml.
+		// The custom configuration is required to run tests using GPU backend.
+		java.io.InputStream inputStream = Thread.currentThread().getContextClassLoader()
+			.getResourceAsStream("my.properties");
 		java.util.Properties properties = new Properties();
-		try {
-			properties.load(inputStream);
+		if(inputStream != null){
+			try {
+				properties.load(inputStream);
+			}
+			catch(IOException e) {
+				e.printStackTrace();
+			}
+			outputBuffering = Boolean.parseBoolean(properties.getProperty("automatedtestbase.outputbuffering"));
+			boolean gpu = Boolean.parseBoolean(properties.getProperty("enableGPU"));
+			TEST_GPU = TEST_GPU || gpu;
+			boolean stats = Boolean.parseBoolean(properties.getProperty("enableStats"));
+			VERBOSE_STATS = VERBOSE_STATS || stats;
 		}
-		catch(IOException e) {
-			e.printStackTrace();
+		else{
+			// If no properties file exists.
+			outputBuffering = false;
+			TEST_GPU = false;
+			VERBOSE_STATS = false;
 		}
-		outputBuffering = Boolean.parseBoolean(properties.getProperty("automatedtestbase.outputbuffering"));
-		TEST_GPU = Boolean.parseBoolean(properties.getProperty("enableGPU"));
-		VERBOSE_STATS = Boolean.parseBoolean(properties.getProperty("enableStats"));
 	}
 
 	// Timestamp before test start.
@@ -336,7 +373,7 @@ public abstract class AutomatedTestBase {
 	protected File getCodegenConfigFile(String parent, CodegenTestType type) {
 		// Instrumentation in this test's output log to show custom configuration file used for template.
 		File tmp = new File(parent, type.getCodgenConfig());
-		if( LOG.isInfoEnabled() )
+		if(LOG.isInfoEnabled())
 			LOG.info("This test case overrides default configuration with " + tmp.getPath());
 		return tmp;
 	}
@@ -380,6 +417,25 @@ public abstract class AutomatedTestBase {
 		return TestUtils.generateTestMatrix(rows, cols, min, max, sparsity, seed);
 	}
 
+	/**
+	 * <p>
+	 * Generates a random matrix with the specified characteristics and returns it as a two dimensional array.
+	 * </p>
+	 *
+	 * @param rows     number of rows
+	 * @param cols     number of columns
+	 * @param min      minimum value
+	 * @param max      maximum value
+	 * @param sparsity sparsity
+	 * @param seed     seed
+	 * @param delta    The minimum value in between values.
+	 * @return two dimensional array containing random matrix
+	 */
+	protected double[][] getRandomMatrix(int rows, int cols, double min, double max, double sparsity, long seed,
+		double delta) {
+		return TestUtils.generateTestMatrix(rows, cols, min, max, sparsity, seed, delta);
+	}
+
 	/**
 	 * <p>
 	 * Generates a random matrix with the specified characteristics which does not contain any zero values and returns
@@ -461,13 +517,7 @@ public abstract class AutomatedTestBase {
 		String completePath = baseDirectory + INPUT_DIR + name + "/in";
 		String completeRPath = baseDirectory + INPUT_DIR + name + ".mtx";
 
-		try {
-			cleanupExistingData(baseDirectory + INPUT_DIR + name, bIncludeR);
-		}
-		catch(IOException e) {
-			e.printStackTrace();
-			throw new RuntimeException(e);
-		}
+		cleanupDir(baseDirectory + INPUT_DIR + name, bIncludeR);
 
 		TestUtils.writeTestMatrix(completePath, matrix);
 		if(bIncludeR) {
@@ -481,6 +531,30 @@ public abstract class AutomatedTestBase {
 		return matrix;
 	}
 
+	protected void writeCSVMatrix(String name, double[][] matrix, boolean header, MatrixCharacteristics mc) {
+		try {
+			final String completePath = baseDirectory + INPUT_DIR + name;
+			final String completeMTDPath = baseDirectory + INPUT_DIR + name + ".mtd";
+			cleanupDir(completePath, false);
+			TestUtils.writeCSV(completePath, matrix, header);
+			final FileFormatProperties ffp = header ? new FileFormatPropertiesCSV(true, ",", false, 0.0, "") : new FileFormatPropertiesCSV();
+			HDFSTool.writeMetaDataFile(completeMTDPath, ValueType.FP64, mc, FileFormat.CSV, ffp);
+		}
+		catch(Exception e) {
+			throw new RuntimeException(e);
+		}
+	}
+
+	protected void cleanupDir(String fullPath, boolean bIncludeR){
+		try {
+			cleanupExistingData(fullPath, bIncludeR);
+		}
+		catch(IOException e) {
+			e.printStackTrace();
+			throw new RuntimeException(e);
+		}
+	}
+
 	protected double[][] writeInputMatrixWithMTD(String name, MatrixBlock matrix, boolean bIncludeR) {
 		double[][] data = DataConverter.convertToDoubleMatrix(matrix);
 		return writeInputMatrixWithMTD(name, data, bIncludeR);
@@ -531,6 +605,99 @@ public abstract class AutomatedTestBase {
 		return matrix;
 	}
 
+	protected void writeInputFederatedWithMTD(String name, MatrixObject fm, PrivacyConstraint privacyConstraint){
+		writeFederatedInputMatrix(name, fm.getFedMapping());
+		MatrixCharacteristics mc = (MatrixCharacteristics)fm.getDataCharacteristics();
+		try {
+			String completeMTDPath = baseDirectory + INPUT_DIR + name + ".mtd";
+			HDFSTool.writeMetaDataFile(completeMTDPath, ValueType.FP64, mc, FileFormat.FEDERATED, privacyConstraint);
+		}
+		catch(IOException e) {
+			e.printStackTrace();
+			throw new RuntimeException(e);
+		}
+
+	}
+
+	protected void writeFederatedInputMatrix(String name, FederationMap fedMap){
+		String completePath = baseDirectory + INPUT_DIR + name;
+		try {
+			cleanupExistingData(baseDirectory + INPUT_DIR + name, false);
+		}
+		catch(IOException e) {
+			e.printStackTrace();
+			throw new RuntimeException(e);
+		}
+
+		ReaderWriterFederated.write(completePath, fedMap);
+		inputDirectories.add(baseDirectory + INPUT_DIR + name);
+	}
+
+	/**
+	 * <p>
+	 * Takes a matrix (double[][]) and writes it in parts locally. Then it creates a federated MatrixObject
+	 * containing the local paths and the given ports. This federated MO is also written to disk with the provided name
+	 * When running federated workers locally on the specified ports this federated Matrix can then be used
+	 * for testing purposes. Just use read on input(name)
+	 * </p>
+	 *
+	 * @param name name of the matrix when writing to disk
+	 * @param matrix two dimensional matrix
+	 * @param numFederatedWorkers the number of federated workers
+	 * @param ports a list of port the length of the number of federated workers
+	 * @param ranges an array containing arrays of length to with the upper and lower bound (rows) for the slices
+	 */
+	protected void rowFederateLocallyAndWriteInputMatrixWithMTD(String name,
+		double[][] matrix, int numFederatedWorkers, List<Integer> ports, double[][] ranges)
+	{
+		// check matrix non empty
+		if(matrix.length == 0 || matrix[0].length == 0)
+			return;
+
+		int nrows = matrix.length;
+		int ncol = matrix[0].length;
+
+		// create federated MatrixObject
+		MatrixObject federatedMatrixObject = new MatrixObject(ValueType.FP64, 
+			Dag.getNextUniqueVarname(Types.DataType.MATRIX));
+		federatedMatrixObject.setMetaData(new MetaDataFormat(
+			new MatrixCharacteristics(nrows, ncol), Types.FileFormat.BINARY));
+
+		// write parts and generate FederationMap
+		List<Pair<FederatedRange, FederatedData>> fedHashMap = new ArrayList<>();
+		for(int i = 0; i < numFederatedWorkers; i++) {
+			double lowerBound = ranges[i][0];
+			double upperBound = ranges[i][1];
+			double examplesForWorkerI = upperBound - lowerBound;
+			String path = name + "_" + (i + 1);
+
+			// write slice
+			writeInputMatrixWithMTD(path, Arrays.copyOfRange(matrix, (int)lowerBound, (int)upperBound),
+				false, new MatrixCharacteristics((long) examplesForWorkerI, ncol,
+				OptimizerUtils.DEFAULT_BLOCKSIZE, (long) examplesForWorkerI * ncol));
+
+			// generate fedmap entry
+			FederatedRange range = new FederatedRange(new long[]{(long) lowerBound, 0}, new long[]{(long) upperBound, ncol});
+			FederatedData data = new FederatedData(DataType.MATRIX, new InetSocketAddress(ports.get(i)), input(path));
+			fedHashMap.add(Pair.of(range, data));
+		}
+		
+		federatedMatrixObject.setFedMapping(new FederationMap(FederationUtils.getNextFedDataID(), fedHashMap));
+		federatedMatrixObject.getFedMapping().setType(FederationMap.FType.ROW);
+
+		writeInputFederatedWithMTD(name, federatedMatrixObject, null);
+	}
+
+	protected double[][] generateBalancedFederatedRowRanges(int numFederatedWorkers, int dataSetSize) {
+		double[][] ranges = new double[numFederatedWorkers][2];
+		double examplesPerWorker = ceil( (double) dataSetSize / (double) numFederatedWorkers);
+		for(int i = 0; i < numFederatedWorkers; i++) {
+			ranges[i][0] = examplesPerWorker * i;
+			ranges[i][1] = Math.min(examplesPerWorker * (i + 1), dataSetSize);
+		}
+		return ranges;
+	}
+
 	/**
 	 * <p>
 	 * Adds a matrix to the input path and writes it to a file.
@@ -690,17 +857,21 @@ public abstract class AutomatedTestBase {
 		expectedFiles.add(baseDirectory + EXPECTED_DIR + cacheDir + name);
 	}
 
-	protected static HashMap<CellIndex, Double> readDMLMatrixFromHDFS(String fileName) {
+	protected static HashMap<CellIndex, Double> readDMLMatrixFromOutputDir(String fileName) {
 		return TestUtils.readDMLMatrixFromHDFS(baseDirectory + OUTPUT_DIR + fileName);
 	}
 
-	public HashMap<CellIndex, Double> readRMatrixFromFS(String fileName) {
-		if( LOG.isInfoEnabled() )
+	protected static HashMap<CellIndex, Double> readDMLMatrixFromExpectedDir(String fileName) {
+		return TestUtils.readDMLMatrixFromHDFS(baseDirectory + EXPECTED_DIR + fileName);
+	}
+	
+	public HashMap<CellIndex, Double> readRMatrixFromExpectedDir(String fileName) {
+		if(LOG.isInfoEnabled())
 			LOG.info("R script out: " + baseDirectory + EXPECTED_DIR + cacheDir + fileName);
 		return TestUtils.readRMatrixFromFS(baseDirectory + EXPECTED_DIR + cacheDir + fileName);
 	}
 
-	protected static HashMap<CellIndex, Double> readDMLScalarFromHDFS(String fileName) {
+	protected static HashMap<CellIndex, Double> readDMLScalarFromOutputDir(String fileName) {
 		return TestUtils.readDMLScalarFromHDFS(baseDirectory + OUTPUT_DIR + fileName);
 	}
 
@@ -742,55 +913,93 @@ public abstract class AutomatedTestBase {
 		return reader.readFrameFromHDFS(strFrameFileName, md.getRows(), md.getCols());
 	}
 
-	public HashMap<CellIndex, Double> readRScalarFromFS(String fileName) {
-		if( LOG.isInfoEnabled() )
+	public HashMap<CellIndex, Double> readRScalarFromExpectedDir(String fileName) {
+		if(LOG.isInfoEnabled())
 			LOG.info("R script out: " + baseDirectory + EXPECTED_DIR + cacheDir + fileName);
 		return TestUtils.readRScalarFromFS(baseDirectory + EXPECTED_DIR + cacheDir + fileName);
 	}
 
 	public static void checkDMLMetaDataFile(String fileName, MatrixCharacteristics mc) {
+		checkDMLMetaDataFile(fileName, mc, false);
+	}
+	
+	public static void checkDMLMetaDataFile(String fileName, MatrixCharacteristics mc, boolean checkBlocksize) {
 		MatrixCharacteristics rmc = readDMLMetaDataFile(fileName);
 		Assert.assertEquals(mc.getRows(), rmc.getRows());
 		Assert.assertEquals(mc.getCols(), rmc.getCols());
+		if( checkBlocksize )
+			Assert.assertEquals(mc.getBlocksize(), rmc.getBlocksize());
 	}
 
 	public static MatrixCharacteristics readDMLMetaDataFile(String fileName) {
 		try {
-			JSONObject meta = getMetaDataJSON(fileName);
-			long rlen = Long.parseLong(meta.get(DataExpression.READROWPARAM).toString());
-			long clen = Long.parseLong(meta.get(DataExpression.READCOLPARAM).toString());
-			return new MatrixCharacteristics(rlen, clen, -1, -1);
+			MetaDataAll meta = getMetaData(fileName);
+			return new MatrixCharacteristics(
+				meta.getDim1(), meta.getDim2(), meta.getBlocksize(), -1);
 		}
 		catch(Exception ex) {
 			throw new RuntimeException(ex);
 		}
 	}
 
-	public static JSONObject getMetaDataJSON(String fileName) {
-		return getMetaDataJSON(fileName, OUTPUT_DIR);
+	public static MetaDataAll getMetaData(String fileName) {
+		return getMetaData(fileName, OUTPUT_DIR);
 	}
 
-	public static JSONObject getMetaDataJSON(String fileName, String outputDir) {
+	public static MetaDataAll getMetaData(String fileName, String outputDir) {
 		String fname = baseDirectory + outputDir + fileName + ".mtd";
-		return new DataExpression().readMetadataFile(fname, false);
-	}
-
-	public static String readDMLMetaDataValue(String fileName, String outputDir, String key) throws JSONException {
-		JSONObject meta = getMetaDataJSON(fileName, outputDir);
-		return meta.get(key).toString();
+		return new MetaDataAll(fname, false, true);
 	}
 
 	/**
-	 * Call readDMLMetaDataValue but fail test in case of JSONException or NullPointerException. 
-	 * @param fileName of metadata file
+	 * Returns the privacy constraint as read from metadata file.
+	 * @param fileName name of file
+	 * @return loaded privacy constraint
+	 * @throws DMLRuntimeException in case of problems with reading the metadata file
+	 */
+	public static PrivacyConstraint getPrivacyConstraintFromMetaData(String fileName, String dir) throws DMLRuntimeException {
+		PrivacyConstraint outputPrivacyConstraint = new PrivacyConstraint();
+		try {
+			MetaDataAll metadata = getMetaData(fileName, dir);
+			Object metaValue;
+			if ( metadata.mtdExists() ){
+				if ( (metaValue = metadata.getPrivacy()) != null){
+					outputPrivacyConstraint.setPrivacyLevel(((PrivacyConstraint) metaValue).getPrivacyLevel());
+				} else {
+					outputPrivacyConstraint.setPrivacyLevel(PrivacyLevel.None);
+				}
+				if ((metaValue = metadata.getFineGrainedPrivacy()) != null ){
+					PrivacyUtils.putFineGrainedConstraintsFromString(outputPrivacyConstraint.getFineGrainedPrivacy(), (String) metaValue);
+				}
+			}
+		} catch (JSONException e){
+			throw new DMLRuntimeException("Exception when reading from meta data file", e);
+		}
+		return outputPrivacyConstraint;
+	}
+
+	public static PrivacyConstraint getPrivacyConstraintFromMetaData(String fileName) throws DMLRuntimeException {
+		return getPrivacyConstraintFromMetaData(fileName, OUTPUT_DIR);
+	}
+
+	public static String readDMLMetaDataPrivacyValue(String fileName, String outputDir, String key) {
+		MetaDataAll meta = getMetaData(fileName, outputDir);
+		return key.equals(DataExpression.FINE_GRAINED_PRIVACY) ? meta.getFineGrainedPrivacy() : meta.getPrivacy().getPrivacyLevel().name();
+	}
+
+	/**
+	 * Call readDMLMetaDataValue but fail test in case of JSONException or NullPointerException.
+	 * 
+	 * @param fileName  of metadata file
 	 * @param outputDir directory of metadata file
-	 * @param key key to find in metadata
+	 * @param key       key to find in metadata
 	 * @return value retrieved from metadata for the given key
 	 */
-	public static String readDMLMetaDataValueCatchException(String fileName, String outputDir, String key){
+	public static String readDMLMetaDataPrivacyValueCatchException(String fileName, String outputDir, String key) {
 		try {
-			return readDMLMetaDataValue(fileName, outputDir, key);
-		} catch (JSONException | NullPointerException e){
+			return readDMLMetaDataPrivacyValue(fileName, outputDir, key);
+		}
+		catch(NullPointerException e) {
 			fail("Privacy constraint not written to output metadata file:\n" + e);
 			return null;
 		}
@@ -798,8 +1007,8 @@ public abstract class AutomatedTestBase {
 
 	public static ValueType readDMLMetaDataValueType(String fileName) {
 		try {
-			JSONObject meta = getMetaDataJSON(fileName);
-			return ValueType.fromExternalString(meta.get(DataExpression.VALUETYPEPARAM).toString());
+			MetaDataAll meta = getMetaData(fileName);
+			return meta.getValueType();
 		}
 		catch(Exception ex) {
 			throw new RuntimeException(ex);
@@ -887,14 +1096,17 @@ public abstract class AutomatedTestBase {
 			String localTemp = curLocalTempDir.getPath();
 			String configContents = configTemplate
 				.replace(createXMLElement(DMLConfig.SCRATCH_SPACE, "scratch_space"),
-					createXMLElement(DMLConfig.SCRATCH_SPACE, localTemp + "/scratch_space"))
+					createXMLElement(DMLConfig.SCRATCH_SPACE, localTemp + "/target/scratch_space"))
 				.replace(createXMLElement(DMLConfig.LOCAL_TMP_DIR, "/tmp/systemds"),
 					createXMLElement(DMLConfig.LOCAL_TMP_DIR, localTemp + "/localtmp"));
 
-			FileUtils.write(getCurConfigFile(), configContents, "UTF-8");
+			if(!disableConfigFile){
 
-			if( LOG.isDebugEnabled() )
-				LOG.debug("This test case will use SystemDS config file %s\n" + getCurConfigFile());
+				FileUtils.write(getCurConfigFile(), configContents, "UTF-8");
+
+				if(LOG.isDebugEnabled())
+					LOG.debug("This test case will use SystemDS config file %s\n" + getCurConfigFile());
+			}
 		}
 		catch(IOException e) {
 			throw new RuntimeException(e);
@@ -908,25 +1120,6 @@ public abstract class AutomatedTestBase {
 		return String.format("<%s>%s</%s>", tagName, value, tagName);
 	}
 
-	/**
-	 * <p>
-	 * Loads a test configuration with its parameters. Adds the output directories to the output list as well as to the
-	 * list of possible comparison files.
-	 * </p>
-	 *
-	 * @param configurationName test configuration name
-	 *
-	 */
-	@Deprecated
-	protected void loadTestConfiguration(String configurationName) {
-		if(!availableTestConfigurations.containsKey(configurationName))
-			fail("test configuration not available: " + configurationName);
-
-		TestConfiguration config = availableTestConfigurations.get(configurationName);
-
-		loadTestConfiguration(config);
-	}
-
 	/**
 	 * Runs an R script, default to the old way
 	 */
@@ -940,7 +1133,8 @@ public abstract class AutomatedTestBase {
 	 */
 	protected void runRScript(boolean newWay) {
 
-		String executionFile = sourceDirectory + selectedTest + ".R";;
+		String executionFile = sourceDirectory + selectedTest + ".R";
+		;
 		if(fullRScriptName != null)
 			executionFile = fullRScriptName;
 
@@ -969,11 +1163,7 @@ public abstract class AutomatedTestBase {
 				"Rscript --default-packages=methods,datasets,graphics,grDevices,stats,utils");
 			// *** END HACK ***
 		}
-
-		if(System.getProperty("os.name").contains("Windows")) {
-			cmd = cmd.replace('/', '\\');
-			executionFile = executionFile.replace('/', '\\');
-		}
+		
 		if(DEBUG) {
 			if(!newWay) { // not sure why have this condition
 				TestUtils.printRScript(executionFile);
@@ -994,7 +1184,7 @@ public abstract class AutomatedTestBase {
 			if(outputFiles != null && outputFiles.length > 0) {
 				expectedFile = new File(expectedDir.getPath() + "/" + outputFiles[0]);
 				if(expectedFile.canRead()) {
-					if( LOG.isInfoEnabled() )
+					if(LOG.isInfoEnabled())
 						LOG.info("Skipping R script cmd: " + cmd);
 					return;
 				}
@@ -1004,8 +1194,18 @@ public abstract class AutomatedTestBase {
 		String outputR;
 		String errorString;
 		try {
+			// if R < 4.0 on Windows is used, the file separator needs to be Windows style
+			if(System.getProperty("os.name").contains("Windows")) {
+				Process r_ver_cmd = Runtime.getRuntime().exec("RScript --version");
+				String r_ver = IOUtils.toString(r_ver_cmd.getErrorStream());
+				if(!r_ver.contains("4.0")) {
+					cmd = cmd.replace('/', '\\');
+					executionFile = executionFile.replace('/', '\\');
+				}
+			}
+			
 			long t0 = System.nanoTime();
-			if( LOG.isInfoEnabled() ) {
+			if(LOG.isInfoEnabled()) {
 				LOG.info("starting R script");
 				LOG.debug("R cmd: " + cmd);
 			}
@@ -1013,10 +1213,6 @@ public abstract class AutomatedTestBase {
 
 			outputR = IOUtils.toString(child.getInputStream());
 			errorString = IOUtils.toString(child.getErrorStream());
-			if( LOG.isTraceEnabled() ) {
-				LOG.trace("Standard Output from R:" + outputR);
-				LOG.trace("Standard Error from R:" + errorString);
-			}
 			
 			//
 			// To give any stream enough time to print all data, otherwise there
@@ -1024,11 +1220,10 @@ public abstract class AutomatedTestBase {
 			// has been printed
 			//
 			child.waitFor();
-
 			try {
 				if(child.exitValue() != 0) {
 					throw new Exception(
-						"ERROR: R has ended irregularly\n" + outputR + "\nscript file: " + executionFile);
+						"ERROR: R has ended irregularly\n" + buildOutputStringR(outputR, errorString) + "\nscript file: " + executionFile);
 				}
 			}
 			catch(IllegalThreadStateException ie) {
@@ -1038,6 +1233,10 @@ public abstract class AutomatedTestBase {
 				child.destroy();
 			}
 
+			if(!outputBuffering) {
+				System.out.println(buildOutputStringR(outputR, errorString));
+			}
+
 			long t1 = System.nanoTime();
 
 			LOG.info("R is finished (in " + ((double) t1 - t0) / 1000000000 + " sec)");
@@ -1063,6 +1262,16 @@ public abstract class AutomatedTestBase {
 		}
 	}
 
+	private static String buildOutputStringR(String standardOut, String standardError){
+		StringBuilder sb = new StringBuilder();
+		sb.append("R Standard output :\n");
+		sb.append(standardOut);
+		sb.append("\nR Standard Error  :\n");
+		sb.append(standardError);
+		sb.append("\n");
+		return sb.toString();
+	}
+
 	/**
 	 * <p>
 	 * Runs a test for which no exception is expected.
@@ -1173,6 +1382,14 @@ public abstract class AutomatedTestBase {
 	protected ByteArrayOutputStream runTest(boolean newWay, boolean exceptionExpected, Class<?> expectedException,
 		String errMessage, int maxSparkInst) {
 
+		String name = "";
+		final StackTraceElement[] ste = Thread.currentThread().getStackTrace();
+		for(int i = 0; i < ste.length; i++) {
+			if(ste[i].getMethodName().equalsIgnoreCase("invoke0"))
+				name = ste[i - 1].getClassName() + "." + ste[i - 1].getMethodName();
+		}
+		LOG.info("Test method name: " + name);
+
 		String executionFile = sourceDirectory + selectedTest + ".dml";
 
 		if(!newWay) {
@@ -1201,7 +1418,7 @@ public abstract class AutomatedTestBase {
 			args.add(executionFile);
 		}
 
-		addProgramIndependentArguments(args);
+		addProgramIndependentArguments(args, programArgs);
 
 		// program-specific parameters
 		if(newWay) {
@@ -1216,15 +1433,15 @@ public abstract class AutomatedTestBase {
 				TestUtils.printDMLScript(fullDMLScriptName);
 			}
 		}
-		
+
 		ByteArrayOutputStream buff = outputBuffering ? new ByteArrayOutputStream() : null;
 		PrintStream old = System.out;
 		if(outputBuffering)
 			System.setOut(new PrintStream(buff));
-		
+
 		try {
 			String[] dmlScriptArgs = args.toArray(new String[args.size()]);
-			if( LOG.isTraceEnabled() )
+			if(LOG.isTraceEnabled())
 				LOG.trace("arguments to DMLScript: " + Arrays.toString(dmlScriptArgs));
 			main(dmlScriptArgs);
 
@@ -1236,7 +1453,7 @@ public abstract class AutomatedTestBase {
 				fail("expected exception which has not been raised: " + expectedException);
 		}
 		catch(Exception | Error e) {
-			if( !outputBuffering )
+			if(!outputBuffering)
 				e.printStackTrace();
 			if(errMessage != null && !errMessage.equals("")) {
 				boolean result = rCompareException(exceptionExpected, errMessage, e, false);
@@ -1248,7 +1465,7 @@ public abstract class AutomatedTestBase {
 				StringBuilder errorMessage = new StringBuilder();
 				errorMessage.append("\nfailed to run script: " + executionFile);
 				errorMessage.append("\nStandard Out:");
-				if( outputBuffering )
+				if(outputBuffering)
 					errorMessage.append("\n" + buff);
 				errorMessage.append("\nStackTrace:");
 				errorMessage.append(getStackTraceString(e, 0));
@@ -1267,15 +1484,13 @@ public abstract class AutomatedTestBase {
 	 * @param args command-line arguments
 	 * @throws IOException if an IOException occurs in the hadoop GenericOptionsParser
 	 */
-	public static void main(String[] args)
-			throws IOException, ParseException, DMLScriptException
-	{
+	public static void main(String[] args) throws IOException, ParseException, DMLScriptException {
 		Configuration conf = new Configuration(ConfigurationManager.getCachedJobConf());
 		String[] otherArgs = new GenericOptionsParser(conf, args).getRemainingArgs();
 		DMLScript.executeScript(conf, otherArgs);
 	}
 
-	private void addProgramIndependentArguments(ArrayList<String> args) {
+	private void addProgramIndependentArguments(ArrayList<String> args, String[] otherArgs) {
 
 		// program-independent parameters
 		args.add("-exec");
@@ -1287,12 +1502,23 @@ public abstract class AutomatedTestBase {
 			args.add("singlenode");
 		else if(rtplatform == ExecMode.SPARK)
 			args.add("spark");
-		else {
+		else
 			throw new RuntimeException("Unknown runtime platform: " + rtplatform);
-		}
+
 		// use optional config file since default under SystemDS/DML
-		args.add("-config");
-		args.add(getCurConfigFile().getPath());
+		boolean configSpecified = false;
+		if(otherArgs != null)
+			for(String  i: otherArgs)
+				if(i.equals("-config")){
+					configSpecified = true;
+					break;
+				}
+
+
+		if(!configSpecified){
+			args.add("-config");
+			args.add(getCurConfigFile().getPath());
+		}
 
 		if(TEST_GPU)
 			args.add("-gpu");
@@ -1317,41 +1543,63 @@ public abstract class AutomatedTestBase {
 	 * @param port Port to use for the JVM
 	 * @return the process associated with the worker.
 	 */
+	@Deprecated
 	protected Process startLocalFedWorker(int port) {
 		Process process = null;
 		String separator = System.getProperty("file.separator");
 		String classpath = System.getProperty("java.class.path");
-		String path = System.getProperty("java.home")
-					+ separator + "bin" + separator + "java";
-		ProcessBuilder processBuilder = new ProcessBuilder(path, "-cp", 
-				classpath, DMLScript.class.getName(), "-w",  Integer.toString(port), "-stats");
+		String path = System.getProperty("java.home") + separator + "bin" + separator + "java";
+		ProcessBuilder processBuilder = new ProcessBuilder(path, "-cp", classpath, DMLScript.class.getName(), "-w",
+			Integer.toString(port), "-stats");
 
-		try{
+		try {
 			process = processBuilder.start();
 			// Give some time to startup the worker.
 			sleep(FED_WORKER_WAIT);
-		} catch (IOException | InterruptedException e){
+		}
+		catch(IOException | InterruptedException e) {
 			e.printStackTrace();
 		}
 		return process;
 	}
 
 	/**
-	 * Start a thread for a worker. This will share the same JVM,
-	 * so all static variables will be shared.!
+	 * Start a thread for a worker. This will share the same JVM, so all static variables will be shared.!
 	 * 
-	 * Also when using the local Fed Worker thread the statistics printing, 
-	 * and clearing from the worker is disabled.
+	 * Also when using the local Fed Worker thread the statistics printing, and clearing from the worker is disabled.
 	 * 
 	 * @param port Port to use
 	 * @return the thread associated with the worker.
 	 */
 	protected Thread startLocalFedWorkerThread(int port) {
+		return startLocalFedWorkerThread(port, null, FED_WORKER_WAIT);
+	}
+
+	protected Thread startLocalFedWorkerThread(int port, String[] otherArgs) {
+		return startLocalFedWorkerThread(port, otherArgs, FED_WORKER_WAIT);
+	}
+
+	/**
+	 * Start a thread for a worker. This will share the same JVM, so all static variables will be shared.!
+	 * 
+	 * Also when using the local Fed Worker thread the statistics printing, and clearing from the worker is disabled.
+	 * 
+	 * @param port Port to use
+	 * @param sleep  The amount of time to wait for the worker startup. in Milliseconds
+	 * @return the thread associated with the worker.
+	 */
+	protected Thread startLocalFedWorkerThread(int port, int sleep) {
+		return startLocalFedWorkerThread(port, null, sleep);
+	}
+	protected Thread startLocalFedWorkerThread(int port, String[] otherArgs, int sleep) {
 		Thread t = null;
 		String[] fedWorkArgs = {"-w", Integer.toString(port)};
 		ArrayList<String> args = new ArrayList<>();
 
-		addProgramIndependentArguments(args);
+		addProgramIndependentArguments(args, otherArgs);
+		
+		if (otherArgs != null)
+			args.addAll(Arrays.stream(otherArgs).collect(Collectors.toList()));
 
 		for(int i = 0; i < fedWorkArgs.length; i++)
 			args.add(fedWorkArgs[i]);
@@ -1369,7 +1617,7 @@ public abstract class AutomatedTestBase {
 				}
 			});
 			t.start();
-			java.util.concurrent.TimeUnit.MILLISECONDS.sleep(FED_WORKER_WAIT);
+			java.util.concurrent.TimeUnit.MILLISECONDS.sleep(sleep);
 		}
 		catch(InterruptedException e) {
 			e.printStackTrace();
@@ -1379,12 +1627,13 @@ public abstract class AutomatedTestBase {
 
 	/**
 	 * Start java worker in same JVM.
+	 * 
 	 * @param args the command line arguments
 	 * @return the thread associated with the process.s
 	 */
 	public static Thread startLocalFedWorkerWithArgs(String[] args) {
 		Thread t = null;
-		
+
 		try {
 			t = new Thread(() -> {
 				try {
@@ -1531,6 +1780,15 @@ public abstract class AutomatedTestBase {
 		}
 	}
 
+	protected void compareResults(double epsilon, String name1, String name2) {
+		for(int i = 0; i < comparisonFiles.length; i++) {
+			HashMap<MatrixValue.CellIndex, Double> expected = TestUtils.readDMLMatrixFromHDFS(comparisonFiles[i]);
+			HashMap<MatrixValue.CellIndex, Double> output = TestUtils.readDMLMatrixFromHDFS(outputDirectories[i]);
+			TestUtils.compareMatrices(expected, output, epsilon, name1, name2);
+
+		}
+	}
+
 	/**
 	 * <p>
 	 * Compares the results of the computation of the frame with the expected ones.
@@ -1620,14 +1878,15 @@ public abstract class AutomatedTestBase {
 	public void tearDown() {
 		LOG.trace("Duration: " + (System.currentTimeMillis() - lTimeBeforeTest) + "ms");
 
-//		assertTrue("expected String did not occur: " + expectedStdOut,
-//			iExpectedStdOutState == 0 || iExpectedStdOutState == 2);
-//		assertTrue("expected String did not occur (stderr): " + expectedStdErr,
-//			iExpectedStdErrState == 0 || iExpectedStdErrState == 2);
-//		assertFalse("unexpected String occurred: " + unexpectedStdOut, iUnexpectedStdOutState == 1);
+		// assertTrue("expected String did not occur: " + expectedStdOut,
+		// iExpectedStdOutState == 0 || iExpectedStdOutState == 2);
+		// assertTrue("expected String did not occur (stderr): " + expectedStdErr,
+		// iExpectedStdErrState == 0 || iExpectedStdErrState == 2);
+		// assertFalse("unexpected String occurred: " + unexpectedStdOut, iUnexpectedStdOutState == 1);
 		TestUtils.displayAssertionBuffer();
 
 		if(!isOutAndExpectedDeletionDisabled()) {
+
 			TestUtils.removeHDFSDirectories(inputDirectories.toArray(new String[inputDirectories.size()]));
 			TestUtils.removeFiles(inputRFiles.toArray(new String[inputRFiles.size()]));
 
@@ -1645,7 +1904,7 @@ public abstract class AutomatedTestBase {
 		TestUtils.clearAssertionInformation();
 	}
 
-	public boolean bufferContainsString(ByteArrayOutputStream buffer, String str){
+	public boolean bufferContainsString(ByteArrayOutputStream buffer, String str) {
 		return Arrays.stream(buffer.toString().split("\n")).anyMatch(x -> x.contains(str));
 	}
 
@@ -1848,6 +2107,31 @@ public abstract class AutomatedTestBase {
 		return false;
 	}
 
+	/**
+	 * Checks if given strings are all in the set of heavy hitters.
+	 * @param str opcodes for which it is checked if all are in the heavy hitters
+	 * @return true if all given strings are in the set of heavy hitters
+	 */
+	protected boolean heavyHittersContainsAllString(String... str){
+		Set<String> heavyHitters = Statistics.getCPHeavyHitterOpCodes();
+		return Arrays.stream(str).allMatch(heavyHitters::contains);
+	}
+
+	/**
+	 * Returns an array of the given opcodes which are not present in the set of heavy hitter opcodes.
+	 * @param opcodes for which it is checked if they are among the heavy hitters
+	 * @return array of opcodes not found in heavy hitters
+	 */
+	protected String[] missingHeavyHitters(String... opcodes){
+		Set<String> heavyHitters = Statistics.getCPHeavyHitterOpCodes();
+		List<String> missingHeavyHitters = new ArrayList<>();
+		for (String opcode : opcodes){
+			if ( !heavyHitters.contains(opcode) )
+				missingHeavyHitters.add(opcode);
+		}
+		return missingHeavyHitters.toArray(new String[0]);
+	}
+
 	protected boolean heavyHittersContainsString(String str, int minCount) {
 		int count = 0;
 		for(String opcode : Statistics.getCPHeavyHitterOpCodes())
@@ -1855,6 +2139,20 @@ public abstract class AutomatedTestBase {
 		return(count >= minCount);
 	}
 
+	protected boolean heavyHittersContainsString(String str, int minCount, long minCallCount) {
+		int count = 0;
+		long callCount = Long.MAX_VALUE;
+		for(String opcode : Statistics.getCPHeavyHitterOpCodes()) {
+			if(opcode.equals(str)) {
+				count++;
+				long tmpCallCount = Statistics.getCPHeavyHitterCount(opcode);
+				if(tmpCallCount < callCount)
+					callCount = tmpCallCount;
+			}
+		}
+		return (count >= minCount && callCount >= minCallCount);
+	}
+
 	protected boolean heavyHittersContainsSubString(String... str) {
 		for(String opcode : Statistics.getCPHeavyHitterOpCodes())
 			for(String s : str)
@@ -1870,6 +2168,20 @@ public abstract class AutomatedTestBase {
 		return(count >= minCount);
 	}
 
+	protected boolean heavyHittersContainsSubString(String str, int minCount, long minCallCount) {
+		int count = 0;
+		long callCount = Long.MAX_VALUE;
+		for(String opcode : Statistics.getCPHeavyHitterOpCodes()) {
+			if(opcode.contains(str)) {
+				count++;
+				long tmpCallCount = Statistics.getCPHeavyHitterCount(opcode);
+				if(tmpCallCount < callCount)
+					callCount = tmpCallCount;
+			}
+		}
+		return (count >= minCount && callCount >= minCallCount);
+	}
+
 	protected boolean checkedPrivacyConstraintsContains(PrivacyLevel... levels) {
 		for(PrivacyLevel level : levels)
 			if(!(CheckedConstraintsLog.getCheckedConstraints().containsKey(level)))
